--------------------------------------------------------------------------------
-- Company: 
-- Engineer:
--
-- Create Date:    11:39:18 01/27/06
-- Design Name:    
-- Module Name:    HTR_LSB - Behavioral
-- Project Name:   
-- Target Device:  
-- Tool versions:  
-- Description:
--
-- Dependencies:
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
-- read
-- address 0x0		GOL_ready & QPLL_locked & QPLL_err & bc_err & neg_edge & run
-- address 0x4		csr
-- address 0x8		i2c_busy & i2c_fail & 	"00" & locked & ttc_status(1) & rx_status(1)
-- address 0xc		i2c_data
-- address 0x10	GOL0 spy data
-- address 0x14	GOL1 spy data
-- address 0x18	GOL2 spy data
-- address 0x1c	ps counter should always be 0x30
-- write
-- address 0x0		Tp_x to GOLx_data mapping scheme number		
-- address 0x4		csr		
-- address 0x8		i2c address register		
-- address 0xc		i2c data register		
-- address 0x10	GOL0 spy data address counter bit 9 downto 2 and resets bit 1 and 0
-- address 0x14	GOL1 spy data address counter bit 9 downto 2 and resets bit 1 and 0
-- address 0x18	GOL2 spy data address counter bit 9 downto 2 and resets bit 1 and 0
-- address 0x40	commands
--						bit0 clears FLASH buffer address counter
--						bit1 takes a snapshot of GOLx_data		
-- address 0x44	FLASH buffer data register		
-- address 0x48	resets		
-- address 0x4c	csr		
--------------------------------------------------------------------------------
library IEEE,mylib;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use mylib.vtx_pack.all;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

entity HTR_SLB is
    Port ( TP_a : in std_logic_vector(35 downto 0);
           TP_b : in std_logic_vector(35 downto 0);
           loc_bus : inout std_logic_vector(7 downto 0);
           loc_add : in std_logic_vector(6 downto 0);
           loc_rw : in std_logic;
           loc_cs : in std_logic;
           loc_ctrl : out std_logic_vector(2 downto 0);
           init_b : out std_logic;
-- loc_spare signals are unused and grounded by CPLD
           loc_spare : in std_logic_vector(7 downto 0);
           ttc_brcst : in std_logic_vector(7 downto 2);
           ttc_brcst_strb : in std_logic;
           ttc_clk_des : in std_logic;
           ttc_bcr : in std_logic;
           rx_bc0_p : in std_logic;
           rx_bc0_n : in std_logic;
           rx_clk_p : in std_logic;
           rx_clk_n : in std_logic;
           GOL_ready : in std_logic_vector(2 downto 0);
           GOL0_data : out std_logic_vector(31 downto 0);
           GOL1_data : out std_logic_vector(31 downto 0);
           GOL2_data : out std_logic_vector(31 downto 0);
           GOL0_flag : out std_logic_vector(1 downto 0);
           GOL1_flag : out std_logic_vector(1 downto 0);
           GOL2_flag : out std_logic_vector(1 downto 0);
           GOL_reset_b : out std_logic_vector(2 downto 0);
           GOL_negedge : out std_logic;
           GOL_cav : out std_logic_vector(2 downto 0);
           GOL_dav : out std_logic_vector(2 downto 0);
           GOL_FF : out std_logic_vector(2 downto 0);
           GOL_SCL : out std_logic_vector(2 downto 0);
           GOL_SDA : inout std_logic;
           QPLL_locked : in std_logic;
           QPLL_err : in std_logic;
           QPLL_reset_b : out std_logic;
           CB_SI : out std_logic;
           CB_EN : out std_logic;
           CB_CK : out std_logic
			  );
end HTR_SLB;

architecture Behavioral of HTR_SLB is
constant version : std_logic_vector(7 downto 0) := x"01";
constant sel_flash : std_logic_vector(6 downto 5) := "01";
constant sel_maddr : std_logic_vector(6 downto 2) := "11000";
COMPONENT m32_1
	PORT(
		d : IN std_logic_vector(31 downto 0);
		s : IN std_logic_vector(4 downto 0);          
		o : OUT std_logic
		);
END COMPONENT;
component ibufgds
    port (i : in std_logic; ib : in std_logic; o : out std_logic);
end component;
component ibufds
    port (i : in std_logic; ib : in std_logic; o : out std_logic);
end component;
component xnor8
    port (o : out std_logic; i0 : in std_logic; i1 : in std_logic;
	       i2 : in std_logic; i3 : in std_logic; i4 : in std_logic;
	       i5 : in std_logic; i6 : in std_logic; i7 : in std_logic
	 );
end component;
signal rx_status : std_logic_vector(7 downto 0);
signal ttc_status : std_logic_vector(7 downto 0);
signal locked : std_logic_vector(1 downto 0);
signal rst_cntr : std_logic_vector(3 downto 0) := x"0";
signal ttc_clk : std_logic := '0';
signal ttc_clk_dcm : std_logic;
signal run : std_logic;
signal rx_clk_in : std_logic;
signal rx_clk : std_logic;
signal rx_clk_dcm : std_logic;
signal rx_clk90 : std_logic;
signal rx_clk90_dcm : std_logic;
signal rx_bc0 : std_logic;
signal got_bc0 : std_logic;
signal TP_a_q : std_logic_vector(35 downto 0);
signal TP_b_q : std_logic_vector(35 downto 0);
signal bc_err : std_logic;
signal s : std_logic_vector(511 downto 0);
signal data : std_logic_vector(71 downto 0);
signal data_a : std_logic_vector(71 downto 0);
signal data_b : std_logic_vector(71 downto 0);
signal GOL0_datap : std_logic_vector(31 downto 0);
signal GOL1_datap : std_logic_vector(31 downto 0);
signal GOL2_datap : std_logic_vector(31 downto 0);
signal addra : std_logic_vector(8 downto 0);
signal addrb : std_logic_vector(8 downto 0);
signal r_data : std_logic_vector(7 downto 0);
signal csr : std_logic_vector(7 downto 0) := x"00";
signal resets : std_logic_vector(7 downto 0) := x"00";
signal cmd : std_logic_vector(7 downto 0) := x"00";
signal rst_dl :std_logic_vector(15 downto 0) := x"0000";
signal test_data : std_logic_vector(7 downto 0) := x"00";
signal rst_dcm :std_logic;
signal psen :std_logic;
signal psdone :std_logic;
signal got_psdone :std_logic;
signal psinc :std_logic;
signal neg_edge :std_logic;
signal calib_going :std_logic;
signal ttc_clk_des_q :std_logic_vector(1 downto 0);
signal ttc_clk_des_q2 :std_logic_vector(1 downto 0);
signal ps :std_logic_vector(6 downto 0);
signal ps_state : std_logic_vector(10 downto 0);
signal ps_cntr :std_logic_vector(5 downto 0);
signal addr :std_logic_vector(13 downto 0);
signal loc_clk : std_logic;
signal loc_clkn : std_logic;
signal i2c_sync : std_logic_vector(1 downto 0);
signal i2c_addr : std_logic_vector(4 downto 0);
signal i2c_data : std_logic_vector(7 downto 0);
signal i2c_wdata : std_logic_vector(7 downto 0);
signal i2c_sr : std_logic_vector(17 downto 0);
signal i2c_phase : std_logic_vector(2 downto 0);
signal i2c_scl : std_logic_vector(2 downto 0);
signal i2c_cntr : std_logic_vector(4 downto 0);
signal clk_div : std_logic_vector(4 downto 0);
signal i2c_id : std_logic;
signal i2c_ec : std_logic;
signal i2c_read : std_logic;
signal i2c_eoc : std_logic;
signal data_cycle : std_logic;
signal i2c_toggle : std_logic;
signal i2c_start : std_logic;
signal i2c_active : std_logic;
signal i2c_done : std_logic;
signal i2c_fail : std_logic;
signal i2c_busy : std_logic;
signal we_flash : std_logic := '0';
signal empty_buffer :std_logic := '1';
signal flash_addra : std_logic_vector(10 downto 0) := "00000000000";
signal flash_addrb : std_logic_vector(10 downto 0) := "00000000000";
signal flash_buf : std_logic_vector(7 downto 0);
signal divider : std_logic_vector(4 downto 0) := "00000";
signal flash_ce : std_logic := '0';
signal rd_buffer : std_logic := '0';
signal clk_ctrl : std_logic_vector(10 downto 0);
signal en_CB_CK : std_logic_vector(11 downto 0);
signal ttc_clkn : std_logic;
signal cmd1_sync : std_logic_vector(1 downto 0);
signal we_gol : std_logic;
signal gol_addrb : std_logic_vector(8 downto 0);
signal gol0_addra : std_logic_vector(10 downto 0);
signal gol1_addra : std_logic_vector(10 downto 0);
signal gol2_addra : std_logic_vector(10 downto 0);
signal gol0_buf : std_logic_vector(7 downto 0);
signal gol1_buf : std_logic_vector(7 downto 0);
signal gol2_buf : std_logic_vector(7 downto 0);
signal status0 : std_logic_vector(7 downto 0);
signal status1 : std_logic_vector(7 downto 0);
signal oslb_reg : std_logic_vector(7 downto 0);
attribute iostandard : string;
attribute iostandard of i_rxclk : label is "lvds_25";
attribute iostandard of i_rxbc0 : label is "lvds_25";
attribute iob : string;
attribute iob of ttc_clk_des_q : signal is "true";
attribute iobdelay : string;
attribute iobdelay of TP_a : signal is "IFD";
attribute iobdelay of TP_b : signal is "IFD";
attribute init : string;
attribute init of rd_buffer : signal is "0";
begin
init_b <= '1';
loc_clk <= loc_spare(5);
process(ttc_clk_des,ttc_status)
begin
-- if clkin stops toggling, reset DCM until clkin comes back
  if(ttc_status(1) = '1' or rx_status(1) = '1')then
    rst_cntr <= x"8";
  elsif(ttc_clk_des'event and ttc_clk_des = '1')then
    if(rst_cntr(3) = '1')then
      rst_cntr <= rst_cntr + 1;
	 end if;
  end if;
end process;
i_dcm_ttc: dcm
   generic map(STARTUP_WAIT => TRUE)  
	port map	
         ( CLKIN   => ttc_clk_des ,
           CLKFB   => ttc_clk ,
           RST     => rst_cntr(3) ,
           DSSEN   => '0' ,
           PSINCDEC=> '0' ,
           PSEN    => '0' ,
           PSCLK   => '0' ,
			  CLK0    => ttc_clk_dcm ,
			  STATUS  => ttc_status ,
			  LOCKED  => locked(0));
bufg_clk: bufg port map(o => ttc_clk, i => ttc_clk_dcm);
-- PCK2111 control, all outputs enabled and using QPLL outputs
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
    if(resets(6) = '1')then
	   clk_ctrl <= "000000" & not csr(2) & '1' & not csr(0) & not csr(1) & csr(3);
	 else
		clk_ctrl <= clk_ctrl(9 downto 0) & '0';
	 end if;
	 en_CB_CK <= en_CB_CK(10 downto 0) & resets(6);
  end if;
end process;
CB_EN <= not resets(6);
CB_SI <= clk_ctrl(10);
ttc_clkn <= not ttc_clk;
i_CB_CK:  fddrrse
	      port map(
			    d0 => '0',
				 d1 => en_CB_CK(11),
				 c0 => ttc_clk,
				 c1 => ttc_clkn,
				 ce => '1',
				 r  => '0',
				 s  => '0',
				 q  => CB_CK);
i_rxclk: ibufgds port map(i => rx_clk_p, ib => rx_clk_n, o => rx_clk_in);
i_rxbc0: ibufds port map(i => rx_bc0_p, ib => rx_bc0_n, o => rx_bc0);
process(rx_clk)
begin
  if(rx_clk'event and rx_clk = '1')then
    if(rx_bc0 = '1')then
	   got_bc0 <= '1';
	 else
	   got_bc0 <= '0';
	 end if;
  end if;
end process;
-- mapping control ROMs, init values are in ucf file
i_g_rom: for i in 0 to 7 generate
  i_rom: RAMB16_S36_S36
		   port map(
			   dia => x"00000000",
				dib => x"00000000",
				dipa => x"0",
				dipb => x"0",
				ena => '1',
				enb => '1',
				wea => '0',
				web => '0',
				ssra => '0',
				ssrb => '0',
				clka => ttc_clk,
				clkb => ttc_clk,
				addra => addra,
				addrb => addrb,
				doa => s(31+i*64 downto i*64),
				dob => s(63+i*64 downto 32+i*64)
				);
end generate;
addrb <= addra(8 downto 1) & '1';
-- multiplexers
i_g0: for i in 0 to 7 generate
  i_g1: for j in 0 to 8 generate
    i_TP_a_mux: m32_1 PORT MAP(
		d => TP_a_q(31 downto 0),
		s => s(4+i*64+j*7 downto i*64+j*7),
		o => data_a(i*9+j));
    i_TP_b_mux: m32_1 PORT MAP(
		d => TP_b_q(31 downto 0),
		s => s(4+i*64+j*7 downto i*64+j*7),
		o => data_b(i*9+j));
	 data(i*9+j) <= not s(6+i*64+j*7) and ((not s(5+i*64+j*7) and data_a(i*9+j)) or (s(5+i*64+j*7) and data_b(i*9+j)));	
  end generate i_g1;
end generate i_g0;
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
    TP_a_q <= TP_a;
    TP_b_q <= TP_b;
	 if(TP_a_q(35 downto 32) /= TP_b_q(35 downto 32))then
	   bc_err <= '1';
    else
	   bc_err <= '0';
	 end if;
--	 GOL0_datap(27 downto 0) <= TP_a_q(35 downto 32) & data(23 downto 0);
--	 GOL1_datap(27 downto 0) <= TP_a_q(35 downto 32) & data(47 downto 24);
--	 GOL2_datap(27 downto 0) <= TP_a_q(35 downto 32) & data(71 downto 48);
  end if;
end process;
GOL0_datap(27 downto 0) <= TP_a_q(35 downto 32) & data(23 downto 0);
GOL1_datap(27 downto 0) <= TP_a_q(35 downto 32) & data(47 downto 24);
GOL2_datap(27 downto 0) <= TP_a_q(35 downto 32) & data(71 downto 48);
g_parity: for i in 0 to 3 generate
  i_GOL0_parity: xnor8 port map(o => GOL0_datap(28+i), i0 => GOL0_datap(i*7), i1 => GOL0_datap(i*7+1),
                 i2 => GOL0_datap(i*7+2), i3 => GOL0_datap(i*7+3), i4 => GOL0_datap(i*7+4), 
                 i5 => GOL0_datap(i*7+5), i6 => GOL0_datap(i*7+6), i7 => '0'); 
  i_GOL1_parity: xnor8 port map(o => GOL1_datap(28+i), i0 => GOL1_datap(i*7), i1 => GOL1_datap(i*7+1),
                 i2 => GOL1_datap(i*7+2), i3 => GOL1_datap(i*7+3), i4 => GOL1_datap(i*7+4), 
                 i5 => GOL1_datap(i*7+5), i6 => GOL1_datap(i*7+6), i7 => '0'); 
  i_GOL2_parity: xnor8 port map(o => GOL2_datap(28+i), i0 => GOL2_datap(i*7), i1 => GOL2_datap(i*7+1),
                 i2 => GOL2_datap(i*7+2), i3 => GOL2_datap(i*7+3), i4 => GOL2_datap(i*7+4), 
                 i5 => GOL2_datap(i*7+5), i6 => GOL2_datap(i*7+6), i7 => '0'); 
end generate;
GOL_cav <= "000";
--process(rx_clk)
--begin
--  if(rx_clk'event and rx_clk = '1')then
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
    if(csr(7) = '1')then
      test_data <= test_data + 1;	 
	   GOL0_data <= x"00" & test_data & x"10" & test_data;
	   GOL1_data <= x"01" & test_data & x"11" & test_data;
	   GOL2_data <= x"02" & test_data & x"12" & test_data;
      GOL_dav <= "111";
	 elsif(run = '1' or csr(6) = '1')then
      GOL0_data <= GOL0_datap;
      GOL1_data <= GOL1_datap;
      GOL2_data <= GOL2_datap;
      GOL_dav <= "111";
	 else
      GOL_dav <= "000";
    end if;
  end if;
end process;
GOL0_flag <= "00";
GOL1_flag <= "00";
GOL2_flag <= "00";
GOL_FF <= "000";
GOL_reset_b <= not resets(5 downto 3);
QPLL_reset_b <= not resets(2);
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
    if(ttc_brcst_strb = '1')then
	   if(ttc_bcr = '1' or ttc_brcst = "101010")then
		  run <= '0';
		elsif(ttc_brcst = "010101")then
		  run <= '1';
		end if;
	 end if;
  end if;
end process;
status0 <= GOL_ready & QPLL_locked & QPLL_err & bc_err & neg_edge & run;
status1 <= i2c_busy & i2c_fail & 	"00" & locked & ttc_status(1) & rx_status(1);
r_data <= flash_buf when rd_buffer = '1' else oslb_reg;
process(loc_add,status0,status1,csr,i2c_data,gol0_buf,gol1_buf,gol2_buf)
begin
  case loc_add(3 downto 0) is
    when "0000" => oslb_reg <= status0;
    when "0001" => oslb_reg <= csr;
    when "0010" => oslb_reg <= status1;
    when "0011" => oslb_reg <= i2c_data;
    when "0100" => oslb_reg <= gol0_buf;
    when "0101" => oslb_reg <= gol1_buf;
    when "0110" => oslb_reg <= gol2_buf;
    when "0111" => oslb_reg <= version;
    when "1000" => oslb_reg <= '0' & ps;
    when "1001" => oslb_reg <= '0' & ps_state(6 downto 0);
    when "1010" => oslb_reg <= "0000" & ps_state(10 downto 7);
    when "1011" => oslb_reg <= x"00";
    when "1100" => oslb_reg <= gol0_addra(7 downto 0);
    when "1101" => oslb_reg <= gol1_addra(7 downto 0);
    when "1110" => oslb_reg <= gol2_addra(7 downto 0);
    when others => oslb_reg <= "00" & gol2_addra(9 downto 8) & gol1_addra(9 downto 8) & gol0_addra(9 downto 8);
  end case;
end process;
loc_bus <= r_data when (loc_cs = '0' and loc_rw = '1' and loc_add(6 downto 5) /= sel_flash) or flash_ce = '1' else (others => 'Z');
process(loc_clk)
begin
  if(loc_clk'event and loc_clk = '1')then
    if(loc_rw = '0' and	loc_add = "0000000")then
	   addra <= loc_bus & '0';
	 end if;
    if(loc_rw = '0' and	loc_add = "0010011")then
	   csr <= loc_bus;
	 end if;
  end if;
end process;
process(loc_clk,rst_dl)
begin
  if(rst_dl(15) = '1')then
    resets <= x"00";
  elsif(loc_clk'event and loc_clk = '1')then
    if(loc_rw = '0' and	loc_add = "0010010")then
	   for i in 1 to 7 loop
	     resets(i) <= loc_bus(i) or loc_bus(0);
		end loop;
      resets(0) <= '1';		
	 end if;
  end if;
end process;
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
	 rst_dl <= rst_dl(14 downto 0) & resets(0);
  end if;
end process;
-- following codes are used to optimize GOL_negedge
rst_dcm <= resets(1) or not locked(0); -- this initiates the calibration
i_dcm_rx: dcm
   generic map(CLKOUT_PHASE_SHIFT => "VARIABLE",
-- starting phase shift after reset
					PHASE_SHIFT => -48,
				   STARTUP_WAIT => TRUE)  
	port map	
         ( CLKIN   => rx_clk_in ,
           CLKFB   => rx_clk ,
           RST     => rst_dcm ,
           DSSEN   => '0' ,
           PSINCDEC=> psinc ,
           PSEN    => psen ,
           PSCLK   => rx_clk ,
			  CLK0    => rx_clk_dcm ,
			  CLK90    => rx_clk90_dcm ,
			  PSDONE  => psdone ,
			  STATUS  => rx_status ,
			  LOCKED  => locked(1));
bufg_rx_clk: bufg port map(o => rx_clk, i => rx_clk_dcm);
bufg_rx_clk90: bufg port map(o => rx_clk90, i => rx_clk90_dcm);
GOL_negedge <= neg_edge;
process(rx_clk)
begin
  if(rx_clk'event and rx_clk = '1')then
    ttc_clk_des_q2(1) <= ttc_clk_des_q(1);
    ttc_clk_des_q(1) <= ttc_clk_des;
	 if(calib_going = '1')then
	   neg_edge <= (not ttc_clk_des_q2(0) and ttc_clk_des_q(0)) or (ttc_clk_des_q2(1) and not ttc_clk_des_q(1));
		ps_state <=  ttc_clk_des_q2 & ttc_clk_des_q & ps;
	 end if;
  end if;
end process;
-- maximum phase shift is 10ns, must use both edges of the clock
process(rx_clk90)
begin
  if(rx_clk90'event and rx_clk90 = '1')then
    ttc_clk_des_q2(0) <= ttc_clk_des_q(0);
    ttc_clk_des_q(0) <= ttc_clk_des;
  end if;
end process;
-- code to automatic align clock edge by tuning phase shift value
process(rx_clk,locked)
begin
  if(locked /= "11")then
	 psen <= '0';
	 got_psdone <= '1'; 
	 ps_cntr <= "100000";
	 calib_going <= '1';
	 psinc <= '1';
	 ps <= "0000000";
  elsif(rx_clk'event and rx_clk = '1')then
	 psen <= ps_cntr(5) and got_psdone;
	 if(psen = '1')then
		if(psinc = '1')then
		  ps <= ps + 1;
		else
		  ps <= ps - 1;
		end if;
	 end if;
	 if((ttc_clk_des_q2(1) /= ttc_clk_des_q(1)) or (ttc_clk_des_q2(0) /= ttc_clk_des_q(0)))then
	   calib_going <= '0';
	 end if;
	 if(calib_going = '0' and ps > "0110000")then
		psinc <= '0';
	 end if;
	 if(psen = '1')then
	   got_psdone <= '0';
	 else
	   got_psdone <= got_psdone or psdone;
	 end if;
	 if(ps_cntr(5) = '1' or (calib_going = '0' and ps = "0110000"))then
	   ps_cntr <= "000000";
	 else
	   ps_cntr <= ps_cntr + 1;
	 end if;
  end if;
end process;
-- I2C interface
process(loc_clk)
begin
  if(loc_clk'event and loc_clk = '1')then
    if(loc_add(6 downto 1) = "000001" and loc_rw = '0' and (loc_add(0) = '1' or loc_bus(7) = '1'))then
	   i2c_toggle <= not i2c_toggle;
	 end if;
    if(loc_add = "0000010" and loc_rw = '0')then
	   i2c_addr <= loc_bus(4 downto 0);
-- there is an error in GOL address wiring
	   i2c_id <= loc_bus(4) or loc_bus(3);
	   i2c_read <= loc_bus(7);
	 end if;
	 if(loc_add = "0000011" and loc_rw = '0')then
	   i2c_wdata <= loc_bus;
	 end if;
  end if;
end process;
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
    i2c_sync <= i2c_sync(0) & i2c_toggle;
	 i2c_start <= i2c_sync(1) xor i2c_sync(0);
    if(i2c_start = '1')then
	   i2c_busy <= '1';
    else
	   i2c_busy <= i2c_busy and not i2c_done;
	 end if;
    if(i2c_start = '1')then
	   i2c_done <= '0';
    elsif(i2c_ec = '1')then
		i2c_done <= i2c_eoc and (data_cycle or i2c_fail);
	 end if;
	 if(i2c_busy = '0')then
	   clk_div <= "00000";
	 else	
      clk_div <= clk_div + 1;
	 end if;
	 if(clk_div = "11111")then
	   i2c_ec <= '1';
	 else
	   i2c_ec <= '0';
	 end if;
  end if;
end process;
-- i2c state machine
-- since this is the only master, no STOP is generated
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
    if(i2c_start = '1')then
		i2c_active <= '0';
		i2c_phase <= "000";
		data_cycle <= '0';
    elsif(i2c_ec = '1')then
	   i2c_active <= i2c_busy;
		i2c_phase(2 downto 1) <= i2c_phase(1 downto 0);
		i2c_phase(0) <= i2c_busy and (i2c_phase(2) or not i2c_active);
		data_cycle <= i2c_active and ((i2c_eoc and i2c_phase(2)) or data_cycle);
	 end if;
  end if;
end process;
-- i2c_fail resets at the beginning of every i2c access
-- if acknoledge from slave is '1', i2c_fail is set and terminates the
-- operation
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
    if(i2c_ec = '1')then
	   if(i2c_busy = '1' and i2c_active = '0')then
		  i2c_fail <= '0';
		elsif(GOL_SDA = '1' and i2c_phase(1) = '1')then
		  if(i2c_cntr = "01001" or
		    (i2c_cntr = "10010" and (data_cycle = '0' or i2c_read = '0')))then
		    i2c_fail <= '1';
		  end if;
		end if;   
	 end if;
  end if;
end process;
GOL_SCL <= i2c_scl;
-- scl generation and i2c_sr shift register 
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
    if(i2c_active = '0')then
      i2c_scl <= "111";
		i2c_sr <= (others => '1');
	   i2c_eoc <= '0';
		i2c_cntr <= "00000";
    elsif(i2c_ec = '1')then
	   if(i2c_addr(4 downto 3) = "00")then
	     i2c_scl(0) <= (not i2c_phase(1) and i2c_scl(0)) or i2c_phase(0) or i2c_eoc;
		end if;
	   if(i2c_addr(4 downto 3) = "01")then
	     i2c_scl(1) <= (not i2c_phase(1) and i2c_scl(1)) or i2c_phase(0) or i2c_eoc;
		end if;
	   if(i2c_addr(4 downto 3) = "10")then
	     i2c_scl(2) <= (not i2c_phase(1) and i2c_scl(2)) or i2c_phase(0) or i2c_eoc;
		end if;
	   if(i2c_phase(0) = '1' and i2c_cntr = "00000")then
		  i2c_sr(17 downto 8) <=  "01000" & i2c_id & '0' & data_cycle & (i2c_read and data_cycle) & '1';
		  if(data_cycle = '0')then
		    i2c_sr(7 downto 0) <= "00000" & i2c_addr(2 downto 0);
		  elsif(i2c_read = '1')then
		    i2c_sr(7 downto 0) <= "11111111"; 
		  else
		    i2c_sr(7 downto 0) <= i2c_wdata(7 downto 0); 
		  end if;
		elsif(i2c_phase(2) = '1')then
		  i2c_sr <= i2c_sr(16 downto 0) & '1';
		end if;
      if(i2c_phase(1) = '1' and (i2c_cntr(4) = '0' or i2c_cntr(1) = '0'))then
	     i2c_data <= i2c_data(6 downto 0) & GOL_SDA;
	   end if;
	   if(i2c_phase(2) = '1')then
	     if(i2c_cntr = "10010" or i2c_fail = '1')then
		    i2c_eoc <= '1';
		  else
		    i2c_eoc <= '0';
		  end if;
		  if(i2c_eoc = '1')then
		    i2c_cntr <= "00000";
		  else
		    i2c_cntr <= i2c_cntr + 1;
		  end if;
		end if;
	 end if;
  end if;
end process;
-- sda and scl outputs
GOL_SDA <= '0' when i2c_eoc = '1' or i2c_sr(17) = '0' else 'Z';
-- FLASH programming code
loc_ctrl(2) <= '1';
loc_ctrl(1) <= not rd_buffer;
loc_ctrl(0) <= not flash_ce;
process(loc_clk)
begin
  if(loc_clk'event and loc_clk = '1')then
    if(loc_add = "0010000" and loc_rw = '0' and loc_bus(0) = '1')then
	   flash_addra(7 downto 0) <= x"00";
	 elsif(loc_add = "0010001" and loc_rw = '0')then
	   flash_addra(7 downto 0) <= flash_addra(7 downto 0) + 1;
	 end if;
    if(loc_add = "0010000" and loc_rw = '0')then
	   cmd <= cmd xor loc_bus;
	 end if;
  end if;
end process;
process(loc_clk,flash_addrb)
begin
  if(flash_addrb(8) = '1')then
	 rd_buffer <= '0';
	 empty_buffer <= '1';
  elsif(loc_clk'event and loc_clk = '1')then
    if(loc_add = "1100010" and loc_rw = '0' and empty_buffer = '0')then
	   rd_buffer <= '1';
	 end if;	
    if(loc_add = "0010001" and loc_rw = '0')then
	   empty_buffer <= '0';
	 end if;	
  end if;
end process;
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
	 if(rd_buffer = '0')then
	   divider <= "00000";
	 else
	   divider <= divider + 1;
	 end if;
	 flash_ce <= divider(4) xor divider(3);
	 if(rd_buffer = '0')then
	   flash_addrb(8 downto 0) <= "000000000";
	 elsif(divider = "11111")then
	   flash_addrb(8 downto 0) <= flash_addrb(8 downto 0) + 1;
	 end if;
  end if;
end process;
flash_addra(10 downto 8) <= "000";
flash_addrb(10 downto 9) <= "00";
we_flash <= '1' when loc_add = "0010001" and loc_rw = '0' else '0';
i_flash_buf: RAMB16_S9_S9
		   port map(
			   dia => loc_bus,
				dib => x"00",
				dipa => "0",
				dipb => "0",
				ena => '1',
				enb => '1',
				wea => we_flash,
				web => '0',
				ssra => '0',
				ssrb => '0',
				clka => loc_clk,
				clkb => ttc_clk,
				addra => flash_addra,
				addrb => flash_addrb,
				dob => flash_buf
				);
-- output monitoring
process(loc_clk)
begin
  if(loc_clk'event and loc_clk = '1')then
    if(loc_rw = '0' and	loc_add = "0000100")then
	   gol0_addra <= '0' & loc_bus & "00";
	 elsif(loc_rw = '1')then
      if(loc_add = "0000100")then
		  gol0_addra(9 downto 0) <= gol0_addra(9 downto 0) + 1;
      end if;		
	 end if;
    if(loc_rw = '0' and	loc_add = "0000101")then
	   gol1_addra <= '0' & loc_bus & "00";
	 elsif(loc_rw = '1')then
      if(loc_add = "0000101")then
		  gol1_addra(9 downto 0) <= gol1_addra(9 downto 0) + 1;
      end if;		
	 end if;
    if(loc_rw = '0' and	loc_add = "0000110")then
	   gol2_addra <= '0' & loc_bus & "00";
	 elsif(loc_rw = '1')then
      if(loc_add = "0000110")then
		  gol2_addra(9 downto 0) <= gol2_addra(9 downto 0) + 1;
      end if;		
	 end if;
  end if;
end process;
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
    cmd1_sync <= cmd1_sync(0) & cmd(1);
	 if(cmd1_sync(1) /= cmd1_sync(0))then
	   we_gol <= '1';
	 elsif(gol_addrb(7 downto 0) = x"ff")then
	   we_gol <= '1';
	 end if;
    if(we_gol = '0')then
	   gol_addrb <= (others => '0');
	 else
	   gol_addrb(7 downto 0) <= gol_addrb(7 downto 0) + 1;
	 end if;
  end if;
end process;
loc_clkn <= not loc_clk;
i_gol0: RAMB16_S9_S36
		   port map(
				dia => x"00",
			   dib => GOL0_datap,
				dipa => "0",
				dipb => x"0",
				ena => '1',
				enb => '1',
				wea => '0',
				web => we_gol,
				ssra => '0',
				ssrb => '0',
				clka => loc_clkn,
				clkb => ttc_clk,
				addra => gol0_addra,
				addrb => gol_addrb,
				doa => gol0_buf
				);
i_gol1: RAMB16_S9_S36
		   port map(
				dia => x"00",
			   dib => GOL1_datap,
				dipa => "0",
				dipb => x"0",
				ena => '1',
				enb => '1',
				wea => '0',
				web => we_gol,
				ssra => '0',
				ssrb => '0',
				clka => loc_clkn,
				clkb => ttc_clk,
				addra => gol1_addra,
				addrb => gol_addrb,
				doa => gol1_buf
				);
i_gol2: RAMB16_S9_S36
		   port map(
				dia => x"00",
			   dib => GOL2_datap,
				dipa => "0",
				dipb => x"0",
				ena => '1',
				enb => '1',
				wea => '0',
				web => we_gol,
				ssra => '0',
				ssrb => '0',
				clka => loc_clkn,
				clkb => ttc_clk,
				addra => gol2_addra,
				addrb => gol_addrb,
				doa => gol2_buf
				);
end Behavioral;
